import React from 'react';
import { Graph, Addon, Path, Cell, Model, Node, Markup } from '@antv/x6';
import { Drawer, Modal, Table, Tooltip } from 'antd';
import type { ColumnsType } from 'antd/es/table';
import AlgoNode from './AlgoNode';
import ToolBar from './ToolBar';
import DatasourceNodeInfo from './DatasourceNodeInfo';
import ScriptNodeInfo from './ScriptNodeInfo';
import OutputNodeInfo from './OutputNodeInfo';
import ContextMenuView from './ContextMenu';
import ReactDOM from 'react-dom';
import { nanoid } from 'nanoid'
import { signNodesSkip } from './utils';
// getNodesSourceAndTarget

// 侧边栏UI组件
const { Stencil } = Addon;

// 节点状态接口
interface NodeStatus {
  id: string;
  status: 'default' | 'success' | 'failed' | 'running';
  type: 'script' | 'datasource' | 'outputsource';
  label?: string;
}

// 数据源节点
const dataSourceNodes = [
  {
    id: '0',
    shape: 'dag-node',
    height: 80,
    width: 180,
    data: {
      label: 'InputSource',
      status: 'default',
      type: 'datasource'
    },
    ports: {
      items: [{ id: 'bottomPort', group: 'bottom' }]
    }
  },
  {
    id: '10',
    shape: 'dag-node',
    height: 80,
    width: 180,
    data: {
      label: 'OutputSource',
      status: 'default',
      type: 'outputsource'
    },
    ports: {
      items: [{ id: 'topPort', group: 'top' }]
    }
  }
];
// 用户自定义脚本节点
const customScriptNodes = [
  {
    id: '5',
    shape: 'dag-node',
    data: {
      label: 'sql',
      status: 'default',
      type: 'script'
    },
    ports: {
      items: [
        // { id: 'topPort', group: 'top' },
        // { id: 'bottomPort', group: 'bottom' }
      ]
    }
  },
  {
    id: '6',
    shape: 'dag-node',
    data: {
      label: 'pyspark',
      status: 'default',
      type: 'script'
    },
    ports: {
      items: [
        // { id: 'topPort', group: 'top' },
        // { id: 'bottomPort', group: 'bottom' }
      ]
    }
  },
  {
    id: '7',
    shape: 'dag-node',
    data: {
      label: 'python',
      status: 'default',
      type: 'script'
    },
    ports: {
      items: [
        // { id: 'topPort', group: 'top' },
        // { id: 'bottomPort', group: 'bottom' }
      ]
    }
  }
];

// 注册算子节点
Graph.registerNode(
  'dag-node',
  {
    inherit: 'react-shape',
    width: 180,
    height: 36,
    component: <AlgoNode />,
    portMarkup: [Markup.getForeignObjectMarkup()],
    ports: {
      groups: {
        top: {
          position: 'top',
          attrs: {
            fo: {
              width: 10,
              height: 10,
              x: -5,
              y: -5,
              magnet: 'true',
            }
          }
        },
        bottom: {
          position: 'bottom',
          attrs: {
            fo: {
              width: 10,
              height: 10,
              x: -5,
              y: -5,
              magnet: 'true',
            }
          }
        }
      }
    }
  },
  true
);

// 注册边
Graph.registerEdge(
  'dag-edge',
  {
    inherit: 'edge',
    attrs: {
      line: {
        stroke: '#C2C8D5',
        strokeWidth: 1,
        targetMarker: 'block'
      }
    }
  },
  true
);

// 注册连接
Graph.registerConnector(
  'algo-connector',
  (s, e) => {
    const offset = 4;
    const deltaY = Math.abs(e.y - s.y);
    const control = Math.floor((deltaY / 3) * 2);

    const v1 = { x: s.x, y: s.y + offset + control };
    const v2 = { x: e.x, y: e.y - offset - control };

    return Path.normalize(
      `M ${s.x} ${s.y}
       L ${s.x} ${s.y + offset}
       C ${v1.x} ${v1.y} ${v2.x} ${v2.y} ${e.x} ${e.y - offset}
       L ${e.x} ${e.y}
      `
    );
  },
  true
);

const columns: ColumnsType<any> = [
  { title: 'pord_order_id', dataIndex: 'pord_order_id', key: 'pord_order_id' },
  {
    title: 'pord_big_categ_name',
    dataIndex: 'pord_big_categ_name',
    key: 'pord_big_categ_name'
  },
  {
    title: 'pord_categ_name',
    dataIndex: 'pord_categ_name',
    key: 'pord_categ_name'
  },
  { title: 'order_date', dataIndex: 'order_date', key: 'order_date' },
  { title: 'qukc_dun', dataIndex: 'qukc_dun', key: 'qukc_dun' },
  { title: 'qukc_price', dataIndex: 'qukc_price', key: 'qukc_price' }
];
const data: any[] = [
  {
    key: '1',
    pord_order_id: 'vdf6ahjdk',
    pord_big_categ_name: '婴儿粉',
    pord_categ_name: '青护',
    order_date: '2020.01.12',
    qukc_dun: 0.0034,
    qukc_price: 0.07821
  },
  {
    key: '2',
    pord_order_id: 'vdf6ahjdk',
    pord_big_categ_name: '婴儿粉',
    pord_categ_name: '青护',
    order_date: '2020.01.12',
    qukc_dun: 0.0034,
    qukc_price: 0.07821
  },
  {
    key: '3',
    pord_order_id: 'vdf6ahjdk',
    pord_big_categ_name: '婴儿粉',
    pord_categ_name: '青护',
    order_date: '2020.01.12',
    qukc_dun: 0.0034,
    qukc_price: 0.07821
  },
  {
    key: '4',
    pord_order_id: 'vdf6ahjdk',
    pord_big_categ_name: '婴儿粉',
    pord_categ_name: '青护',
    order_date: '2020.01.12',
    qukc_dun: 0.0034,
    qukc_price: 0.07821
  }
];
export default class Dag extends React.Component<any, any> {
  // 容器DIV
  private container: any;
  // 侧边栏UI容器
  private stencilContainer: any;
  // 构造函数
  constructor(props: any) {
    super(props);
    this.state = {
      nodeInfoVisible: false,
      dagGraph: null,
      dagId: '',
      selectedNodeData: {},
      contextMenu: null,
      contextMenuNode: null,
      isOpenTable: false,
      isOpenLog: false
    };
  }

  // 展示node节点信息
  showNodeInfo = () => {
    this.setState({ nodeInfoVisible: true });
  };

  // 隐藏node节点信息
  hideNodeInfo = () => {
    this.setState({ nodeInfoVisible: false });
  };

  // 右键菜单
  handleContextMenu = (event: React.MouseEvent) => {
    event.preventDefault();
    this.setState({
      contextMenu:
        this.state.contextMenu === null
          ? {
              mouseX: event.clientX + 2,
              mouseY: event.clientY - 6
            }
          : null
    });
  };

  // 执行此处
  handleNodeRun = () => {
    const menuNode = this.state.contextMenuNode
    const {script_nodes, script_edges} = signNodesSkip(this.state.dagGraph,menuNode, 'handle_run')
    console.log('script_nodes:', script_nodes);
    console.log('script_edges:', script_edges);
    const data = menuNode.getData() as NodeStatus;
    menuNode.setData({
      ...data,
      status: 'running'
    });
    setTimeout(() => {
      menuNode.setData({
        ...data,
        status: 'success'
      });
    }, 3000);
    this.setState({ contextMenu: null });
  };

  // 执行到此处
  handleRunEndNode = () => {
    const menuNode = this.state.contextMenuNode
    const {script_nodes, script_edges} = signNodesSkip(this.state.dagGraph,menuNode, 'handle_run_end')
    console.log('script_nodes:', script_nodes);
    console.log('script_edges:', script_edges);
    this.setState({ contextMenu: null });
  };
  // 从此处开始执行
  handleRunBeginNode = () => {
    const menuNode = this.state.contextMenuNode
    const {script_nodes, script_edges} = signNodesSkip(this.state.dagGraph,menuNode, 'handle_run_begin')
    console.log('script_nodes:', script_nodes);
    console.log('script_edges:', script_edges);
    this.setState({ contextMenu: null });
  };

  // 关闭右键菜单
  handleContextMenuClose = () => {
    this.setState({ contextMenu: null });
  };

  handleOpenTable = () => {
    this.setState({ contextMenu: null, isOpenTable: true });
  };

  handleOpenLog = () => {
    this.setState({ contextMenu: null, isOpenLog: true });
  };

  // 挂载
  componentDidMount(): void {
    // 创建图
    const graph: Graph = new Graph({
      // 容器
      container: this.container,
      // 格子
      // grid: true,
      background: {
        color: '#FBFCFF'
      },
      // 拖动
      panning: {
        enabled: true,
        eventTypes: ['leftMouseDown', 'mouseWheel']
      },
      // 放大缩小
      mousewheel: {
        enabled: true,
        modifiers: 'ctrl',
        factor: 1.1,
        maxScale: 1.5,
        minScale: 0.5
      },
      // 交互触发高亮
      highlighting: {
        magnetAdsorbed: {
          name: 'stroke',
          args: {
            attrs: {
              fill: '#fff',
              stroke: '#31d0c6',
              strokeWidth: 4
            }
          }
        }
      },
      // 连接
      connecting: {
        snap: true,
        allowBlank: false,
        allowLoop: false,
        highlight: true,
        connector: 'algo-connector',
        connectionPoint: 'anchor',
        anchor: 'center',
        // 验证
        validateMagnet({ magnet }) {
          return magnet.getAttribute('port-group') !== 'top';

        },
        validateConnection({ targetPort, sourceCell, targetCell, targetMagnet }) {
          if (!targetCell) return false
          const incomingEdges = this.getIncomingEdges(targetCell)
          const existPorts = incomingEdges?.map(edge => {
            return edge.getTargetPortId()
          })
          const jugeExist = existPorts?.indexOf(targetPort as any)
          return targetMagnet?.getAttribute('port-group') === 'top' && (jugeExist === undefined || jugeExist === -1);
        },
        createEdge() {
          return graph.createEdge({
            shape: 'dag-edge',
            attrs: {
              line: {
                strokeDasharray: '5 5'
              }
            },
            zIndex: -1
          });
        }
      },
      // 选择
      selecting: {
        enabled: true,
        multiple: true,
        rubberEdge: true,
        rubberNode: true,
        modifiers: 'shift',
        rubberband: true
      },
      // 快捷键
      keyboard: {
        enabled: true,
        global: true
      },
      onPortRendered(args) {
        const selectors = args.contentSelectors
        const port = args.port
        const container = selectors && selectors.foContent
        if (container) {
          ReactDOM.render(
            (
              <Tooltip title={port.id}>
                <div className="my-port" />
              </Tooltip>
            ) as any,
            container as HTMLElement,
          )
        }
      },
    });

    // 删除节点
    graph.bindKey(['delete', 'backspace'], () => {
      graph.getSelectedCells().forEach((item: any) => {
        item.remove();
      });
    });

    // 监听边连接
    graph.on('edge:connected', ({ edge }) => {
      edge.attr({
        line: {
          strokeDasharray: ''
        }
      });
    });

    // 监听状态改变
    graph.on('node:change:data', ({ node }) => {
      const edges = graph.getIncomingEdges(node);
      const { status } = node.getData() as NodeStatus;
      edges?.forEach(edge => {
        if (status === 'running') {
          edge.attr('line/strokeDasharray', 5);
          edge.attr('line/style/animation', 'running-line 30s infinite linear');
        } else {
          edge.attr('line/strokeDasharray', '');
          edge.attr('line/style/animation', '');
        }
      });
    });

    // 监听节点被双击
    graph.on(
      'node:dblclick',
      (args: { cell: Cell; node: Node; options: Model.SetOptions }) => {
        const nodeData = args.cell.data;
        this.setState({ selectedNodeData: nodeData });
        this.showNodeInfo();
      }
    );

    //监听右键
    graph.on('node:contextmenu', ({ node, e }) => {
      if (node.data.type === 'script') {
        this.setState({ contextMenuNode: node });
        this.handleContextMenu(e);
      }
    });

    const text = this.props.context.current;
    text.ready.then(() => {
      const dagJson: any = text.model.toJSON();
      if (dagJson?.graph) {
        graph.fromJSON(dagJson.graph);
        this.setState({ dagId: dagJson?.dag_id || nanoid().replace('-', 'a').replace('_', 'b')})
      } else {
        this.setState({ dagId: nanoid().replace('-', 'a').replace('_', 'b')})
      }
    });

    // 内容居中
    graph.centerContent();

    // 创建侧边栏
    const stencil = new Stencil({
      title: '算子检索',
      target: graph,
      // 搜索
      search(cell, keyword) {
        return cell.data.label.indexOf(keyword) !== -1;
      },
      placeholder: '搜索算子',
      notFoundText: '未找到相关算子',
      scaled: true,
      // animation: true,
      // 侧边栏项宽高
      stencilGraphWidth: 200,
      stencilGraphHeight: 200,
      layoutOptions: {
        columns: 1,
        dx: 55,
        rowHeight: 'compact'
      },
      // 侧边栏项目分组
      groups: [
        {
          name: 'dataSource',
          title: '数据源'
        },
        {
          name: 'dataHandle',
          title: '数据处理',
          collapsed: true
        },
        {
          name: 'otherScript',
          title: '其它',
          collapsed: true
        },
        {
          name: 'customScript',
          title: '自定义脚本'
        }
      ],
      // 拖拽事件 处理节点数据
      getDropNode(draggingNode: Node): Node {
        const newNode = draggingNode.clone();
        initNodeData(newNode);
        return newNode;
      }
    });

    // 添加到侧边栏容器内
    this.stencilContainer.appendChild(stencil.container);

    // 将可拖拽项加载到侧边栏中
    stencil.load(dataSourceNodes, 'dataSource');
    stencil.load(customScriptNodes, 'customScript');
    stencil.resizeGroup('dataSource', {
      width: 200,
      height: dataSourceNodes.length * 100
    });
    stencil.resizeGroup('customScript', {
      width: 200,
      height: customScriptNodes.length * 60
    });

    // 设置图
    this.setState({ dagGraph: graph });
  }

  // 卸载
  componentWillUnmount(): void {
    this.state.dagGraph.dispose();
  }
  // 容器
  refContainer = (container: HTMLDivElement) => {
    this.container = container;
  };

  // 侧边栏
  refStencil = (container: HTMLDivElement) => {
    this.stencilContainer = container;
  };

  // 最后输出模板
  render() {
    return (
      <div className="app">
        <Modal
          title="查看结果表"
          open={this.state.isOpenTable}
          onCancel={() => this.setState({ isOpenTable: false })}
          footer={null}
          width={'60%'}
        >
          <div className="table_toolbar">
            <span className="toolbar_label">总数</span>
            <span>{data.length}</span>
            <span className="toolbar_label">列数</span>
            <span>{columns.length}</span>
          </div>
          <Table
            columns={columns}
            dataSource={data}
            pagination={{ pageSize: 20 }}
            scroll={{ y: 450 }}
          />
        </Modal>
        <Modal
          title="查看执行日志"
          open={this.state.isOpenLog}
          onCancel={() => this.setState({ isOpenLog: false })}
          footer={null}
          width={'60%'}
          bodyStyle={{ padding: 0 }}
        >
          <pre className="log">暂无数据</pre>
        </Modal>
        {/* 侧边栏 */}
        <div className="app-stencil" ref={this.refStencil} />
        {/* 容器 */}
        <div className="app-content">
          <div className="app-graph" ref={this.refContainer} />
          <ContextMenuView
            contextStatus={this.state.contextMenu}
            contextMenuNode={this.state.contextMenuNode}
            handleClose={this.handleContextMenuClose}
            handleRunBeginNode={this.handleRunBeginNode}
            handleRunEndNode={this.handleRunEndNode}
            handleRun={this.handleNodeRun}
            handleOpenTable={this.handleOpenTable}
            handleOpenLog={this.handleOpenLog}
          />
          {/* 节点信息 */}
          <Drawer
            title={this.state.selectedNodeData?.label || 'NodeTitle'}
            placement="right"
            closable={false}
            destroyOnClose={true}
            onClose={this.hideNodeInfo}
            open={this.state.nodeInfoVisible}
            getContainer={false}
            style={{ position: 'absolute' }}
          >
            {/* 数据源节点 */}
            {this.state.selectedNodeData?.type === 'datasource' && (
              <DatasourceNodeInfo nodeInfo={this.state.selectedNodeData} />
            )}
            {/* 脚本节点 */}
            {this.state.selectedNodeData?.type === 'script' && (
              <ScriptNodeInfo nodeInfo={this.state.selectedNodeData} graph={this.state.dagGraph} />
            )}
            {/* 输出源节点 */}
            {this.state.selectedNodeData?.type === 'outputsource' && (
              <OutputNodeInfo nodeInfo={this.state.selectedNodeData} />
            )}
          </Drawer>
          {/* 工具栏 */}
          <ToolBar
            className="app-tool"
            graph={this.state.dagGraph}
            saveGraph={this.props.saveGraph}
            dagId={this.state.dagId}
          />
        </div>
      </div>
    );
  }
}

// 初始化节点数据
function initNodeData(newNode: Node<Node.Properties>) {
  // 节点id
  newNode.data.id = newNode.id;
  newNode.data.nodeId = nanoid().replace('-', 'a').replace('_', 'b')
  switch (newNode.data.type) {
    case 'datasource':
      newNode.data.inputSource = undefined;
      newNode.data.dataTableId = undefined;
      newNode.data.dataTableName = undefined;
      newNode.data.nodeName = undefined;
      break;
    case 'script':
      newNode.data.paramText = undefined;
      newNode.data.nodeName = undefined;
      newNode.data.scriptText = '';
      newNode.data.outputData = [];
      newNode.data.inputNumber = 0;
      newNode.data.packageData = undefined;
      newNode.data.status = 'undone';
      break;
    case 'outputsource':
      newNode.data.nodeName = undefined;
      newNode.data.outputSource = undefined;
    default:
  }
}
